from .error import EmptyLineError


class BDFParser:
    
    BEGINBULK = "BEGIN BULK"
    ENDDATA = "ENDDATA"
    MAXLINELENGTH = 80
    FIELDWIDTH = 8
    FIELDSPERLINE = 10
    FIELDSPERBODY = 8
    
    def __init__(self, bdf_str):
        # bdf_str = self.expand_tabs(bdf_str)
        self.bds = self.ignore_enddata(bdf_str)
        self.lines = self.bds.split("\n")
        self.remove_comments()
        self.line_idx = 0
        self.cards = []
        
    def current_line(self):
        return self.lines[self.line_idx]
    
    def increment_line(self, i=1):
        self.line_idx += i
        return self.current_line()
    
    def ignore_enddata(self, bdf_str):
        
        try:
            enddata_i = bdf_str.index(self.ENDDATA)
        except ValueError:
            return bdf_str
        else:
            return bdf_str[:enddata_i]
            
    def is_comment(self, line):
        line = line.lstrip()
        return not line or line[0] == "$"
        
    def remove_comments(self):
        comments_i = [
            i 
            for i in range(len(self.lines))
            if self.is_comment(self.lines[i])
        ]
        for i in reversed(comments_i):
            del self.lines[i]

    def parse_header(self):

        beginbulk_i = None
        for line_idx, line in enumerate(self.lines):
            if self.BEGINBULK in line:
                beginbulk_i = line_idx
                break

        if beginbulk_i is None:
            return ""
        else:
            self.line_idx = beginbulk_i + 1
            return "\n".join(self.lines[:beginbulk_i])
    
    def is_line_free(self, line):
        return "," in line
    
    def parse_fields(self, fields):
        fieldsperline = self.FIELDSPERLINE
        fieldsperbody = self.FIELDSPERBODY
        numfields = len(fields)
        head, body, tail = None, [], None
        if numfields == 0:
            raise EmptyLineError()
        if numfields > 0:
            head = fields[0]
        if numfields > 1:
            body = fields[1:]
        if numfields == fieldsperline:
            tail = body.pop(-1)
        # missing fields are blank fields
        nummissing  = fieldsperbody - len(body)
        if nummissing > 0:
            body.extend(["" for _ in range(nummissing)])
        return head, body, tail
    
    def parse_line_free(self, line):
        fields = line.rstrip(",").split(",")
        return self.parse_fields(fields)
    
    def parse_line_fixed(self, line):
        fieldwidth = self.FIELDWIDTH
        length = len(line)
        fields = [line[i:i+fieldwidth]
                  for i in range(0, length, fieldwidth)]
        return self.parse_fields(fields)
    
    def parse_line(self, line):
        
        if self.is_line_free(line):
            return self.parse_line_free(line)

        else:
            return self.parse_line_fixed(line)
        
    def parse_card(self):
        
        line = self.current_line()
        name, fields, tail = self.parse_line(line)
        
        while True:
            
            try:
                next_line = self.increment_line()
            except IndexError:
                break

            next_head, next_fields, next_tail = self.parse_line(next_line)
            
            if not tail:
                next_head = next_head.strip()
                if next_head and not ("+" in next_head):
                    break

            tail = next_tail
            fields.extend(next_fields)

        # pop trailing blank fields
        for i in reversed(range(len(fields))):
            if not fields[i].strip():
                del fields[i]
            else:
                break

        return name, fields
            
    def endofbdf(self):
        return self.line_idx == len(self.lines)
        
    def parse(self):
        
        header = self.parse_header()
        
        cards = []
        
        while not self.endofbdf():
            
            card = self.parse_card()
            cards.append(card)
            
        return header, cards


"""
Defines functions for single precision 16 character field writing.
"""
import sys
from typing import List, Union, Optional, Any
from numpy import float32, isnan  # type: ignore

from ..utils.numpy_utils import integer_types
from .cards.utils import wipe_empty_fields
from .field_writer_8 import set_blank_if_default

def set_string16_blank_if_default(value: Any, default: Any) -> str:
    """helper method for writing BDFs"""
    val = set_blank_if_default(value, default)
    if val is None:
        return '                '
    return '%16s' % val

def print_scientific_16(value: float) -> str:
    """
    Prints a value in 16-character scientific notation.
    This is a sub-method and shouldnt typically be called

    .. seealso:: print_float_16 for a better method
    """
    if value == 0.0:
        return '%16s' % '0.'

    python_value = '%16.14e' % value  # -1.e-2
    svalue, sexponent = python_value.strip().split('e')
    exponent = int(sexponent)  # removes 0s

    if abs(value) < 1.:
        sign = '-'
    else:
        sign = '+'

    # the exponent will be added later...
    sexp2 = str(exponent).strip('-+')
    value2 = float(svalue)

    # the plus 1 is for the sign
    len_sexp = len(sexp2) + 1
    leftover = 16 - len_sexp

    if value < 0:
        fmt = "%%1.%sf" % (leftover - 3)
    else:
        fmt = "%%1.%sf" % (leftover - 2)

    svalue3 = fmt % value2
    svalue4 = svalue3.strip('0')
    field = "%16s" % (svalue4 + sign + sexp2)
    return field


def print_float_16(value: float) -> str:
    """
    Prints a float in nastran 16-character width syntax
    using the highest precision possbile.
    .. seealso:: print_float_8
    """
    if isnan(value):
        return '                '
    elif value == 0.0:
        return '%16s' % '0.'
    elif value > 0.:  # positive, not perfect...
        if value < 5e-16:
            field = print_scientific_16(value)
            return field
        elif value < 0.001:
            field = print_scientific_16(value)
            field2 = "%16.15f" % value  # small value
            field2 = field2.strip('0 ')

            field1 = field.replace('-', 'e-')

            if field2 == '.':
                return print_scientific_16(value)
            if len(field2) <= 16 and float(field1) == float(field2):
                field = field2
                field = field.strip(' 0')
            return '%16s' % field
        #elif value < 0.1:
            #field = "%16.15f" % value
        elif value < 1.:
            field = "%16.15f" % value
        elif value < 10.:
            field = "%16.14f" % value
        elif value < 100.:
            field = "%16.13f" % value   # 10 < x < 100
        elif value < 1000.:
            field = "%16.12f" % value
        elif value < 10000.:
            field = "%16.11f" % value
        elif value < 100000.:
            field = "%16.10f" % value
        elif value < 1000000.:
            field = "%16.9f" % value
        elif value < 10000000.:
            field = "%16.8f" % value
        elif value < 100000000.:
            field = "%16.7f" % value
        elif value < 1000000000.:
            field = "%16.6f" % value
        elif value < 10000000000.:
            field = "%16.5f" % value
        elif value < 100000000000.:
            field = "%16.4f" % value
        elif value < 1000000000000.:
            field = "%16.3f" % value
        elif value < 10000000000000.:
            field = "%16.2f" % value
        elif value < 100000000000000.:
            field = "%16.1f" % value
        else:  # big value  # 123456789012345.
            field = "%16.1f" % value
            if field.index('.') < 16:
                field = '%16.1f' % (round(value))
                field = field[0:16]  # drop off the .1f
                #assert '.' != field[0], field
            else:
                field = print_scientific_16(value)
            return field
    else:
        if value > -5e-15:
            field = print_scientific_16(value)
            return field
        elif value > -0.01:  # -0.001
            field = print_scientific_16(value)
            field2 = "%16.14f" % value  # small value
            field2 = field2.strip('0 ')

            # get rid of the first minus sign, add it on afterwards
            field1 = '-' + field.strip(' 0-').replace('-', 'e-')

            if len(field2) <= 16 and float(field1) == float(field2):
                field = field2.rstrip(' 0')
                field = field.replace('-0.', '-.')
            return '%16s' % field
        #elif value > -0.1:
            # -0.01 >x>-0.1...should be 5 (maybe scientific...)
            #field = "%16.14f" % value
            #field = field.replace('-0.', '-.')
        elif value > -1.:
            # -0.1  >x>-1.....should be 6, but the baseline 0 is kept...
            field = "%16.14f" % value
            field = field.replace('-0.', '-.')
        elif value > -10.:
            field = "%16.13f" % value  # -1    >x>-10
        elif value > -100.:
            field = "%16.12f" % value  #       -1 > x >      -10
        elif value > -1000.:
            field = "%16.11f" % value  #      -10 > x >     -100
        elif value > -10000.:
            field = "%16.10f" % value  #     -100 > x >    -1000
        elif value > -100000.:
            field = "%16.9f" % value   #    -1000 > x >   -10000
        elif value > -1000000.:
            field = "%16.8f" % value   #  -10,000 > x > -100,000
        elif value > -10000000.:
            field = "%16.7f" % value   #           -100,000 > x >          -1,000,000
        elif value > -100000000.:
            field = "%16.6f" % value   #         -1,000,000 > x >         -10,000,000
        elif value > -1000000000.:
            field = "%16.5f" % value   #        -10,000,000 > x >        -100,000,000
        elif value > -10000000000.:
            field = "%16.4f" % value   #       -100,000,000 > x >      -1,000,000,000
        elif value > -100000000000.:
            field = "%16.3f" % value   #     -1,000,000,000 > x >     -10,000,000,000
        elif value > -1000000000000.:
            field = "%16.2f" % value   #    -10,000,000,000 > x >    -100,000,000,000
        elif value > -10000000000000.:
            field = "%16.1f" % value   #   -100,000,000,000 > x >  -1,000,000,000,000
        else:
            field = "%16.1f" % value
            try:
                ifield = field.index('.')
            except ValueError:
                print('field = %s' % field)
                raise
            if ifield < 16:
                field = '%15s.' % (int(round(value, 0)))
                #assert '.' != field[0], field
            else:
                field = print_scientific_16(value)
            return field
    field = field.strip(' 0')
    field = '%16s' % field

    #assert len(field) == 16, ('value=%r field=%r is not 16 characters '
    #                          'long, its %s' % (value, field, len(field)))
    return field


def print_field_16(value):
    # type: (Optional[Union[int, float, str]]) -> str
    """
    Prints a 16-character width field

    Parameters
    ----------
    value : int / float / str / None
        the value to print

    Returns
    -------
    field : str
        an 16-character string
    """
    if isinstance(value, integer_types):
        field = "%16s" % value
    elif isinstance(value, (float, float32)):
        field = print_float_16(value)
    elif value is None:
        field = "                "
    else:
        field = "%16s" % value
    if len(field) != 16:
        msg = 'field=%r is not 16 characters long...rawValue=%r' % (field, value)
        raise RuntimeError(msg)
    return field


def print_card_16(fields, wipe_fields=True):
    # type: (List[Optional[Union[int, float, str]]], bool) -> str
    """
    Prints a nastran-style card with 16-character width fields.

    Parameters
    ----------
    fields : List[varies]
        all the fields in the BDF card (no trailing Nones)
    wipe_fields : bool; default=True
        some cards (e.g. PBEAM) have ending fields
        that need to be there, others cannot have them.

    Returns
    -------
    card : str
        string representation of the card in small field format

    .. note:: An internal field value of None or '' will be treated as
              a blank field
    .. note:: A large field format follows the  8-16-16-16-16-8 = 80
              format where the first 8 is the card name or
              blank (continuation).  The last 8-character field indicates
              an optional continuation, but because it's a left-justified
              unneccessary field, print_card doesnt use it.

    .. code-block:: python

       >>> fields = ['DUMMY', 1, 2, 3, None, 4, 5, 6, 7, 8.]
       >>> print_card_16(fields)
       DUMMY*                 1               2               3
       *                      4               5               6               7
       *                     8.
       *

    """
    if wipe_fields:
        fields = wipe_empty_fields(fields)
    nfields_main = len(fields) - 1  # chop off the card name
    nbdf_lines = nfields_main // 8
    if nfields_main % 8 != 0:
        nbdf_lines += 1
        nextra_fields = 8 * nbdf_lines - nfields_main
        fields += [None] * nextra_fields

    try:
        out = '%-8s' % (fields[0] + '*')
    except:
        print("ERROR!  fields=%s" % fields)
        sys.stdout.flush()
        raise

    for i in range(1, len(fields)):
        field = fields[i]
        try:
            out += print_field_16(field)
        except:
            print("bad fields = %s" % fields)
            raise
        if i % 4 == 0:  # allow 1+4 fields per line
            out = out.rstrip(' ')
            if out[-1] == '\n':  # empty line
                out += '*'
            out += '\n*       '
    out = out.rstrip(' *')  # removes one continuation star
    if not out.endswith('\n'):
        out += '\n'
    return out



"""Unit test package for bulkdata."""
import os
import sys
 
TEST_DIR = os.path.dirname(os.path.abspath(__file__))
# PROJECT_DIR = os.path.abspath(os.path.join(TEST_DIR, os.pardir))
# sys.path.insert(0, PROJECT_DIR)

BDF_DIR = TEST_DIR + "/bdf"
EXPECT_DIR = TEST_DIR + "/expect"




#!/usr/bin/env python

"""Tests for `bulkdata.card` module."""

import pytest
from collections.abc import Sequence

from bulkdata.card import Card#, CardType


@pytest.fixture 
def fields():
    return {
        "id": 99,
        "string": "helloworld",
        "integers": [0, 1, 2],
        "reals": [3.3, 6.6, 9.9]
    }


@pytest.fixture
def card_str():
    return """\
HELLO   99      helloworld      0       3.3     1       6.6     2       +0      
+0      9.9
"""


def test_card_blank():

    size = 16
    card = Card("BLANK   ", size=size)
    assert card.name == "BLANK"
    assert len(card.fields) == size
    # format object strips trailing blank fields during write
    expect = """\
BLANK                                                                   +0      
+0
"""
    assert card.dumps() == expect


def test_card_delitem():

    card = Card("DELETE  ")
    card.append("deletethis", fieldspan=2)
    del card[1]
    del card[0]
    expect = """\
DELETE
"""
    assert card.dumps() == expect


def test_card_strip():

    card = Card("STRIP", size=8)
    card[4] = "stophere"
    card.strip()
    print(card.dumps())
    expect = """\
STRIP                                   stophere
"""
    assert card.dumps() == expect


def test_card_setmultifieldvalue():

    large_string = "the-answer-is-42"
    large_number = 1234567891012345

    card = Card(size=2)
    card[:] = large_string
    got_string = "".join(card[:])
    assert got_string == large_string
    
    card[:] = large_number
    got_number = "".join([str(num) for num in card[:]]) # pylint: disable=not-an-iterable
    assert int(got_number) == large_number


def test_card_size_set(fields, card_str):

    numfields = 1 + 2 + len(fields["integers"]) + len(fields["reals"])
    card = Card("HELLO", size=numfields)

    assert len(card.fields) == numfields

    card[0] = fields["id"]
    card[1:3] = fields["string"]
    card[3::2] = fields["integers"]
    card[4::2] = fields["reals"]

    assert len(card.fields) == numfields
    assert card[0] == fields["id"]
    assert "".join(card[1:3]) == fields["string"]
    assert card[3::2] == fields["integers"]
    assert card[4::2] == fields["reals"]

    assert card.dumps() == card_str


def test_card_resize_set(fields, card_str):

    numfields = 1 + 2 + len(fields["integers"]) + len(fields["reals"])
    card = Card("HELLO")
    card.resize(numfields)

    assert len(card.fields) == numfields

    card[0] = fields["id"]
    card[1:3] = fields["string"]
    card[3::2] = fields["integers"]
    card[4::2] = fields["reals"]

    assert len(card.fields) == numfields
    assert card[0] == fields["id"]
    assert "".join(card[1:3]) == fields["string"]
    assert card[3::2] == fields["integers"]
    assert card[4::2] == fields["reals"]

    assert card.dumps() == card_str


def test_card_oversize_set_strip(fields, card_str):

    numfields = 1 + 2 + len(fields["integers"]) + len(fields["reals"])
    size = 1000
    card = Card("HELLO", size=size)

    assert len(card.fields) == size

    card[0] = fields["id"]
    card[1:3] = fields["string"]
    card[3::2] = fields["integers"]
    card[4::2] = fields["reals"]

    card.strip()

    assert len(card.fields) == numfields
    assert card[0] == fields["id"]
    assert "".join(card[1:3]) == fields["string"]
    assert card[3::2] == fields["integers"]
    assert card[4::2] == fields["reals"]

    assert card.dumps() == card_str


def test_card_append():

    card = Card("TEST")
    for value in [1, 2, 3, 4, 5, 6, 7, 8]:
        card.append(value)
    for value in [1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8]:
        card.append(value)
    for value in ["a", "b", "c", "d", "e", "f", "g","h"]:
        card.append(value)

    expect = """\
TEST    1       2       3       4       5       6       7       8       +0      
+0      1.1     2.2     3.3     4.4     5.5     6.6     7.7     8.8     +1      
+1      a       b       c       d       e       f       g       h
"""
    assert card.dumps() == expect


def test_card_extend():

    card = Card("TEST")
    card.extend([1, 2, 3, 4, 5, 6, 7, 8])
    card.extend([1.1, 2.2, 3.3, 4.4, 5.5, 6.6, 7.7, 8.8])
    card.extend(["a", "b", "c", "d", "e", "f", "g","h"])

    expect = """\
TEST    1       2       3       4       5       6       7       8       +0      
+0      1.1     2.2     3.3     4.4     5.5     6.6     7.7     8.8     +1      
+1      a       b       c       d       e       f       g       h
"""
    assert card.dumps() == expect


def test_card_contains():

    values = ["one", 1, 1.0]

    card = Card("TEST")
    card.extend(values)

    for val in values:
        assert val in card

    assert not "two" in card


def test_card_incremental_set_append(fields, card_str):

    card = Card("HELLO")
    card.append(fields["id"])
    card.append(fields["string"], 2)
    for integer, real in zip(fields["integers"], fields["reals"]):
        card.append(integer)
        card.append(real)

    assert card.dumps() == card_str


def test_card_incremental_set_extend(fields, card_str):

    card = Card("HELLO")
    card.extend([fields["id"], "hellowor", "ld"])
    for integer, real in zip(fields["integers"], fields["reals"]):
        card.extend([integer, real])

    assert card.dumps() == card_str


def test_load_card_modify(fields, card_str):

    card = Card.loads(card_str)
    assert card.dumps() == card_str
    
    for value in fields.values():
        if isinstance(value, str):
            assert value[:8] in card
        elif isinstance(value, Sequence):
            for each in value:
                assert each in card
        else:
            assert value in card

    card.name = "GOODBYE"
    card[0] = 100
    card[[1, 2]] = "short"
    card[3::2] = [3, 4, 5]
    card[4::2] = [1.1, 2.2, 3.3]

    modified_card_str = """\
GOODBYE 100     short           3       1.1     4       2.2     5       +0      
+0      3.3
"""

    assert card.dumps() == modified_card_str


def test_card_get_large_string():

    fieldspan = 6
    large_string = "the-answer-to-life-the-universe-and-everything"

    card = Card("LONGSTR")
    card.append(large_string, fieldspan=fieldspan)

    got_string = card.get_large(slice(0, fieldspan, 1))
    assert got_string == large_string

    got_string = card.get_large(slice(0, None, 1))
    assert got_string == large_string

    index = list(range(fieldspan))
    got_string = card.get_large(index)
    assert got_string == large_string

    got_string = card.get_large(0)
    assert got_string == "the-answ"


def test_card_get_large_number():

    large_number =  1000000000000000

    card = Card("LONGNUM")
    card.append(large_number, fieldspan=2)

    got_number = card.get_large(slice(0, 2, 1))
    assert got_number == large_number

    got_number = card.get_large(slice(0, None, 1))
    assert got_number == large_number

    index = list(range(2))
    got_number = card.get_large(index)
    assert got_number == large_number

    got_number = card.get_large(0)
    assert got_number == 10000000


def test_card_freeformat(fields):

    numfields = 1 + 2 + len(fields["integers"]) + len(fields["reals"])
    card = Card("HELLO", size=numfields)

    card[0] = fields["id"]
    card[1:3] = fields["string"]
    card[3::2] = fields["integers"]
    card[4::2] = fields["reals"]

    card_str = """\
HELLO,99,hellowor,ld,0,3.3,1,6.6,2,+0
+0,9.9
"""
    assert card.dumps("free") == card_str


def test_card_sparse():

    card_str = """\
SPARSE  test
        test
        test
"""

    card = Card("SPARSE")

    line = ["test"] + [None for i in range(7)]

    for _ in range(3):
        card.extend(line)

    card2 = Card.loads(card_str)
    assert card.dumps() == card2.dumps()

def test_card_set_incomplete_overwrite():

    card = Card("OVERWRT")
    card.extend([-1 for _ in range(8)])
    card.extend(["isstring" for _ in range(8)])
    card.append("done")

    card[:8] = list(range(7))
    card[8:16] = "shorter"

    card_str = """\
OVERWRT 0       1       2       3       4       5       6               +0      
+0      shorter                                                         +1      
+1      done
"""
    print(card.dumps("fixed"))
    assert card.dumps("fixed") == card_str


"""Defines functions for single precision 8 character field writing."""
import sys
from typing import List, Union, Any
from numpy import float32, isnan


def set_string8_blank_if_default(value: Any, default: Any) -> str:
    """helper method for writing BDFs"""
    val = set_blank_if_default(value, default)
    if val is None:
        return '        '
    return '%8s' % val


def is_same(value1: Any, value2: Any) -> bool:
    """
    Checks to see if 2 values are the same

    .. note:: this method is used by almost every card when printing
    """
    if isinstance(value1, str) or value1 is None:
        return value1 == value2
    if value1 == value2:
        return True
    return False


def set_blank_if_default(value: Any, default: Any) -> Union[int, float, str, None]:
    """
    Used when setting the output data of a card to clear default values

    Parameters
    ----------
    value : int/float/str
        the field value the may be set to None (blank)
        if value=default, the default value for the field
    default : int/float/str
        the default value

    .. note:: this method is used by almost every card when printing
    """
    if isinstance(value, (float, float32)) and isnan(value):
        return None
    return None if is_same(value, default) else value


def set_default_if_blank(value: Any, default: Any) -> Union[int, float, str]:
    """
    Used when initializing a card and the default value isn't set
    Used on PBARL"""
    return default if value is None or value == '' else value


def print_scientific_8(value: float) -> str:
    """
    Prints a value in 8-character scientific notation.
    This is a sub-method and shouldnt typically be called

    Notes
    -----
    print_float_8 : a better float printing method

    """
    if value == 0.0:
        return '%8s' % '0.'

    python_value = '%8.11e' % value
    svalue, sexponent = python_value.strip().split('e')
    exponent = int(sexponent)  # removes 0s

    sign = '-' if abs(value) < 1. else '+'

    # the exponent will be added later...
    exp2 = str(exponent).strip('-+')
    value2 = float(svalue)

    leftover = 5 - len(exp2)

    if value < 0:
        fmt = "%%1.%sf" % (leftover - 1)
    else:
        fmt = "%%1.%sf" % leftover

    svalue3 = fmt % value2
    svalue4 = svalue3.strip('0')
    field = "%8s" % (svalue4 + sign + exp2)
    return field


def print_float_8(value: float) -> str:
    """
    Prints a float in nastran 8-character width syntax using the
    highest precision possbile.

    """
    if isnan(value):
        return '        '
    elif value == 0.0:
        return '%8s' % '0.'
    elif value > 0.:  # positive, not perfect...
        if value < 5e-8:
            field = print_scientific_8(value)
            return field
        elif value < 0.001:
            field = print_scientific_8(value)
            field2 = "%8.7f" % value  # small value
            field2 = field2.strip('0 ')

            field1 = field.replace('-', 'e-')

            if field2 == '.':
                return print_scientific_8(value)
            if len(field2) <= 8 and float(field1) == float(field2):
                field = field2
                field = field.strip(' 0')
        #elif value < 0.1:
            #field = "%8.7f" % value
        elif value < 1.:
            field = "%8.7f" % value  # same as before...
        elif value < 10.:
            field = "%8.6f" % value
        elif value < 100.:
            field = "%8.5f" % value
        elif value < 1000.:
            field = "%8.4f" % value
        elif value < 10000.:
            field = "%8.3f" % value
        elif value < 100000.:
            field = "%8.2f" % value
        elif value < 1000000.:
            field = "%8.1f" % value
        else:  # big value
            field = "%8.1f" % value
            if field.index('.') < 8:
                field = '%8.1f' % round(value)
                field = field[0:8]
                #assert '.' != field[0], field
            else:
                field = print_scientific_8(value)
            return field
    else:
        if value > -5e-7:
            field = print_scientific_8(value)
            return field
        elif value > -0.01:  # -0.001
            field = print_scientific_8(value)
            field2 = "%8.6f" % value  # small value
            field2 = field2.strip('0 ')

            # get rid of the first minus sign, add it on afterwards
            field1 = '-' + field.strip(' 0-').replace('-', 'e-')

            if len(field2) <= 8 and float(field1) == float(field2):
                field = field2.rstrip(' 0')
                field = field.replace('-0.', '-.')

        #elif value > -0.1:
            # -0.01 >x>-0.1...should be 5 (maybe scientific...)
            #field = "%8.6f" % value
            #field = field.replace('-0.', '-.')
        elif value > -1.:
            # -0.1  >x>-1.....should be 6, but the baseline 0 is kept...
            field = "%8.6f" % value
            field = field.replace('-0.', '-.')
        elif value > -10.:
            field = "%8.5f" % value   # -1    >x>-10
        elif value > -100.:
            field = "%8.4f" % value   # -10   >x>-100
        elif value > -1000.:
            field = "%8.3f" % value   # -100  >x>-1000
        elif value > -10000.:
            field = "%8.2f" % value   # -1000 >x>-10000
        elif value > -100000.:
            field = "%8.1f" % value   # -10000>x>-100000
        elif value <= -999999.5:
            field = print_scientific_8(value)
            return field
        else:
            field = "%8.1f" % value
            try:
                ifield = field.index('.')
            except ValueError:
                raise ValueError('error printing float; cant find decimal; field=%r value=%s' % (
                    field, value))
            if ifield < 8:
                field = '%7s.' % int(round(value, 0))
                #assert '.' != field[0], field
            else:
                field = print_scientific_8(value)
            return field
    field = field.strip(' 0')
    field = '%8s' % field

    #assert len(field) == 8, ('value=%r field=%r is not 8 characters '
    #                         'long, its %s' % (value, field, len(field)))
    return field


#def print_float_or_int_8(value: Union[int, float]) - str:
    #"""
    #Prints a 8-character width field

    #Parameters
    #----------
    #value : int/float
        #the value to print

    #Returns
    #-------
    #field : str
        #an 8-character string
    #"""
    #if isinstance(value, (float, float32)):
        #field = print_float_8(value)
    #elif isinstance(value, int):
        #field = "%8i" % value
    #else:
        #msg = 'Invalid Type:  value=%r type=%s' % (value, type(value))
        #raise TypeError(msg)
    #return field


def print_field_8(value: Union[int, float, str, None]) -> str:
    """
    Prints a 8-character width field

    Parameters
    ----------
    value : int/float/str
        the value to print

    Returns
    -------
    field : str
        an 8-character string

    """
    if isinstance(value, int):
        field = "%8i" % value
    elif isinstance(value, (float, float32)):
        field = print_float_8(value)
    elif value is None:
        field = "        "
    else:
        field = "%8s" % value
    if len(field) != 8:
        msg = 'field=%r is not 8 characters long...raw_value=%r' % (field, value)
        raise RuntimeError(msg)
    return field


def print_card_8(fields: List[Union[int, float, str, None]]) -> str:
    """
    Prints a nastran-style card with 8-character width fields.

    Parameters
    ----------
    fields : List[int/float/str/None]
        all the fields in the BDF card (no trailing Nones)

    Returns
    -------
    card : str
        string representation of the card in small field format

    .. note:: An internal field value of None or '' will be treated as
              a blank field
    .. note:: A small field format follows the  8-8-8-8-8-8-8-8 = 80
              format where the first 8 is the card name or
              blank (continuation).  The last 8-character field indicates
              an optional continuation, but because it's a left-justified
              unneccessary field, print_card doesnt use it.

    .. code-block:: python

       >>> fields = ['DUMMY', 1, 2, 3, None, 4, 5, 6, 7, 8.]
       >>> print_card_8(fields)
       DUMMY          1       2       3               4       5       6       7
       DUMMY          1       2       3               4       5       6       7
                     8.

    """
    try:
        out = '%-8s' % fields[0]
    except:
        print("ERROR!  fields=%s" % fields)
        sys.stdout.flush()
        raise

    for i in range(1, len(fields)):
        field = fields[i]
        try:
            out += print_field_8(field)
        except:
            print("bad fields = %s" % fields)
            raise
        if i % 8 == 0:  # allow 1+8 fields per line
            out = out.rstrip(' ')
            if out[-1] == '\n':  # empty line
                out += '+'
            out += '\n        '
    out = out.rstrip(' \n+') + '\n'  # removes blank lines at the end of cards
    return out


def print_int_card(fields: List[Union[int]]) -> str:
    """
    Prints a nastran-style card with 8-character width fields.
    All fields (other than the first field) must be integers.
    This is used to speed up SET cards.

    Parameters
    ----------
    fields : List[int/float/str/None]
      The list of fields to write to a nastran card.

    .. warning::
      Blanks are not allowed!
      Floats and strings are not allowed.

    .. code-block:: python

       fields = ['SET', 1, 2, 3, 4, 5, 6, ..., n]

    """
    try:
        out = '%-8s' % fields[0]
    except:
        print("ERROR!  fields=%s" % fields)
        sys.stdout.flush()
        raise

    for i in range(1, len(fields)):
        field = fields[i]
        try:
            out += "%8i" % field  # balks if you have None or string fields
        except:
            print("bad fields = %s" % fields)
            raise
        if i % 8 == 0:  # allow 1+8 fields per line
            out = out.rstrip(' ')
            out += '\n        '
    out = out.rstrip(' \n+') + '\n'  # removes blank lines at the end of cards
    return out


def print_int_card_blocks(fields_blocks: List[Any]) -> str:
    """
    Prints a nastran-style card with 8-character width fields.
    All fields other than the card name must be written in "block" format.
    This is used to speed up SET cards.

    Parameters
    ----------
    fields_blocks : List[int]
      The fields written in "block" notation.

    Returns
    -------
    msg : str
        the field blocks as a 8-character width Nastran card

    .. note:: Blanks are allowed in the False block.

    .. code-block:: python

       fields_blocks = [
           'SET1',
           [['a', 1.0, 3], False], # these are not all integers
           [[1, 2, 3], True],      # these are all integers
       ]
       msg = print_int_card_blocks(fields_blocks)
       print(msg)
       >>> 'SET1           a      1.       3       1       2       3\n'

    """
    card_name = fields_blocks[0]
    try:
        out = '%-8s' % card_name
    except:
        print("ERROR!  fields_blocks=%s" % fields_blocks)
        sys.stdout.flush()
        raise

    i = 0
    for block in fields_blocks[1:]:
        (fields, is_all_ints) = block
        if is_all_ints is True:
            for field in fields:
                out += "%8i" % field
                i += 1
                if i == 8:  # allow 1+8 fields per line
                    out += '\n        '
                    i = 0
        elif is_all_ints is False:
            for field in fields:
                out += print_field_8(field)
                i += 1
                if i == 8:  # allow 1+8 fields per line
                    out += '\n        '
                    i = 0
        else:
            raise SyntaxError('is_all_ints must be a boolean.  is_all_ints=%r' % is_all_ints)
    out = out.rstrip(' \n') + '\n'  # removes blank lines at the end of cards
    return out



"""The :mod:`~bulkdata.deck` module provides the 
:class:`~bulkdata.deck.Deck` class.
"""

from collections.abc import Sequence

from .card import Card
from .field import Field, write_field
from .util import islist, repr_list
from .parse import BDFParser


class Deck():
    """:class:`~bulkdata.deck.Deck` class allows the user
    to load and update bulk data files, loading the 
    bulk data cards into :class:`~bulkdata.card.Card`
    objects.

    :param cards: initialize the deck with these cards, 
                    defaults to ``None``.
    :param header: the header, which is prepended to the bulk
                    data section when dumping the deck, 
                    defaults to ``None``.
    """
    
    def __init__(self, cards=None, header=None):
        self._cards = cards or []
        self.header = header or ""
        
    def append(self, card):
        """Append a card to the deck.

        :param card: The card to append
        """
        self._cards.append(card)
        
    def extend(self, cards):
        """Extend deck cards with sequence of cards.

        :param cards: The sequence of cards
        """
        self._cards.extend(cards)
    
    def _iter(self, value):
        if islist(value):
            for each in value:
                yield each
        else:
            yield value
            
    def _iter_index_value(self, dict_):
        index = dict_["index"]
        value = dict_["value"]
        if islist(value):
            if not islist(index):
                raise TypeError("value is type {} but index is type {}"
                                .format(type(value), type(index)))
            for each_index, each_value in zip(index, value):
                yield each_index, each_value
        else:
            yield index, value
        
    def _matches_name(self, filter, card):
        try:
            return card.name == filter["name"]
        except KeyError:
            return True
        
    def _matches_fields(self, filter, card):
        filter_fields = filter.get("fields")
        if filter_fields:
            match = True
            for index, value in self._iter_index_value(filter_fields):
                try:
                    match *= (card.fields[index].value == value)
                except IndexError:
                    return False
            return match
        else:
            return True
        
    def _matches_contains(self, filter, card):
        filter_contains = filter.get("contains")
        if filter_contains:
            filter_contains = [
                value
                for value in self._iter(filter_contains)
            ]
            # loop through fields, popping off matches
            for field in card.fields:
                try:
                    filter_contains.remove(field.value)
                except ValueError:
                    continue
            # match if all `contains` values found
            match = len(filter_contains) == 0
            return match
        else:
            return True
        
    def _matches(self, filter, card):
        match = True
        match *= self._matches_name(filter, card)
        match *= self._matches_fields(filter, card)
        match *= self._matches_contains(filter, card)
        return match
    
    def _enumerate_find(self, filter=None):
        filter = filter or {}
        if filter:
            for i, card in enumerate(self._cards):
                if self._matches(filter, card):
                    yield i, card
        else:
            for i, card in enumerate(self._cards):
                yield i, card
    
    def _enumerate_find_one(self, filter=None):
        filter = filter or {}
        if filter:
            for i, card in enumerate(self._cards):
                if self._matches(filter, card):
                    return i, card
            return None, None
        else:
            try:
                return 0, self._cards[0]
            except IndexError:
                return None, None

    def _normalize_filter(self, filter):
        if filter is None:
            return {}
        elif isinstance(filter, str):
            return {"name": filter}
        else:
            return filter

    def find(self, filter=None):
        """Find cards matching the query denoted by *filter*.

        :type filter: dict, str
        :param filter: Specifies which cards to find.
        :return: A generator object iterating through every card
                 matching the filter.

        If *filter* is a ``dict``, there are three keywords that may
        be used.

        * **name**, ``str``: Filter for cards with matching *name*.

        * **fields**, ``dict``: Given an "index" and "field" member, filter
          cards where the field(s) at *index* index, has/have
          value *field*.

        * **contains**: Filter for cards containing a field that matches
          the *contains* value, or any *contains* value if *contains*
          is a list.

        The following code block is an example of using ``find`` with
        a filter dict:

        .. code-block:: python

            filter_ = {
                # name is ASET1
                "name": "ASET1",
                
                "fields": {
                    
                    # first field
                    "index": 0,
                    
                    # with value 3
                    "value": 3
                },
                
                # contains values 1 and "THRU"
                "contains": [1, "THRU"] 
            }

            card = next(deck.find(filter_))
            print(card)

        .. code-block:: none

            ASET1   3       1       THRU    8

        If *filter* is a ``str``, the filter will match cards
        with name matching *filter*.

        .. code-block:: python

            card = next(deck.find("AERO"))
            print(card)

        .. code-block:: none

            AERO    3       1.3     100.    .00001  1       -1

        """
        filter = self._normalize_filter(filter)
        for _, card in self._enumerate_find(filter):
            yield card

    def find_one(self, filter=None):
        """Find the first card matching the query denoted by *filter*.

        :type filter: dict, str
        :param filter: Specifies which card to find.
        :return: The first matching card, or ``None`` if no match
                 is found.
        """
        filter = self._normalize_filter(filter)
        _, card = self._enumerate_find_one(filter)
        return card
    
    def replace(self, filter, card):
        """Replace cards matching the query denoted by
        *filter* with *card*.

        :type filter: dict, str
        :param filter: Specifies which cards to replace.
        :param card: The replacement card
        """
        filter = self._normalize_filter(filter)
        for i, _ in self._enumerate_find(filter):
            self._cards[i] = card

    def replace_one(self, filter, card):
        """Replace the first card matching the query denoted by
        *filter* with *card*.

        :type filter: dict, str
        :param filter: Specifies which card to replace.
        :param card: The replacement card
        :return: The replacement card, or ``None`` if no match
                 is found.
        """
        filter = self._normalize_filter(filter)
        i, _ = self._enumerate_find_one(filter)
        if i:
            self._cards[i] = card
            return card
        else:
            return None
            
    def _update_card(self, card, update):
        for index, value in self._iter_index_value(update):
            card[index] = value
    
    def update(self, filter, update):
        """Update cards matching the query denoted by *filter*
        with changes denoted by *update* dict, which has
        the same keyword options as the *fields* dict in *filter*.

        .. note:: 
        
            Avoid this function as there is no current tutorial,
            has not been well tested, and does not appear to add any
            functionality not achieved otherwise.

        :type filter: dict, str
        :param filter: Specifies which cards to replace.
        :param card: The replacement card
        """
        filter = self._normalize_filter(filter)
        for i, _ in self._enumerate_find(filter):
            self._update_card(self._cards[i], update)
    
    def delete(self, filter=None):
        """Delete cards matching the query denoted by
        *filter*.

        :type filter: dict, str
        :param filter: Specifies which cards to delete.
        :return: The number of cards deleted.
        """
        filter = self._normalize_filter(filter)
        delete_i = [i for i, _ in self._enumerate_find(filter)]
        for i in reversed(delete_i):
            del self._cards[i]
        return len(delete_i)
            
    def _get_card_by_index(self, index):
        return self._cards[index]
            
    def _get_cards_by_indexes(self, indexes):
        return [self._cards[i] for i in indexes]

    def _get_cards_by_name(self, name):
        return list(self.find({"name": name}))
    
    def _get_cards_by_slice(self, slice_):
        return list(self._cards[slice_])
            
    def __getitem__(self, key):
        """Get card(s) in the deck.

        :type key: int, slice, list, str
        :param key: The indexing key denoting where to get 
                    the card(s)
        :return: The card(s)

        If *key* is of type ``str``, this method
        returns all cards with name *key*.
        """
        if isinstance(key, int):
            return self._get_card_by_index(key)
        elif isinstance(key, slice):
            return self._get_cards_by_slice(key)
        elif isinstance(key, str):
            return self._get_cards_by_name(key)
        elif isinstance(key, Sequence):
            return self._get_cards_by_indexes(key)
        else:
            raise TypeError(key, type(key))
            
    def _set_card_by_index(self, index, card):
        self._cards[index] = card
            
    def _set_cards_by_indexes(self, indexes, cards):
        for i, card in zip(indexes, cards):
            self._cards[i] = card
    
    def _set_cards_by_slice(self, slice_, cards):
        steps = range(slice_.start, slice_.stop. slice_.step)
        for i, card in zip(steps, cards):
            self._cards[i] = card
            
    def __setitem__(self, key, value):
        """Set card(s) in the deck.

        :type key: int, slice, list
        :param key: The indexing key denoting where to set 
                    the card(s)
        :param value: The card(s) to set
        """
        if isinstance(key, int):
            return self._set_card_by_index(key, value)
        elif isinstance(key, slice):
            return self._set_cards_by_slice(key, value)
        elif islist(key):
            return self._set_cards_by_indexes(key, value)
        else:
            raise TypeError(key, type(key))

    @classmethod
    def loads(cls, deck_str):
        """Load :class:`~bulkdata.deck.Deck` object from a
        bulk data string.

        :param deck_str: the bulk data string
        :return: The loaded :class:`~bulkdata.deck.Deck` object
        """
        cards = []
        header, card_tuples = BDFParser(deck_str).parse()

        for name, fields in card_tuples:
            card = Card(name)
            fields = [Field(field_val) for field_val in fields]
            card.set_raw_fields(fields)
            cards.append(card)
        obj = cls(cards, header)

        if obj.find_one({"name": None}):
            raise Warning("Loaded cards with no name. This usually "
                          "implies there was an error parsing the bdf file.")
        
        return obj

    @classmethod
    def load(cls, fp):
        """Load :class:`~bulkdata.deck.Deck` object from a
        bulk data file object.

        :param fp: the bulk data file object
        :return: The loaded :class:`~bulkdata.deck.Deck` object
        """
        return cls.loads(fp.read())

    def dumps(self, format="fixed"):
        """Dump the deck to a bulk data string.

        :param format: the desired format, can be one of: 
                       ["free", "fixed"], defaults to "fixed"
        :return: The bulk data string
        """
        bulk = "".join([card.dumps(format) for card in self.cards])
        if self.header:
            return self.header + "\nBEGIN BULK\n" + bulk + "ENDDATA"
        else:
            return bulk

    def dump(self, fp, format="fixed"):
        """Dump the deck to a bulk data file.

        :param fp: the bulk data file object
        :param format: the desired format, can be one of: 
                       ["free", "fixed"], defaults to "fixed"
        """
        return fp.write(self.dumps(format=format))

    def __str__(self):
        """Dump the deck to as a bulk data string with default
        format
        
        :return: The bulk data string
        """
        return self.dumps()

    def __len__(self):
        """Return number of cards in the deck.
        """
        return self.cards.__len__()

    def __iter__(self):
        """Iterate through the cards in the deck.
        """
        return self.cards.__iter__()

    def __bool__(self):
        """Return ``True`` if the deck contains any cards,
        ``False`` otherwise.
        """ 
        return bool(self.cards)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, 
                               repr_list(self.cards))

    @property
    def cards(self):
        """The deck cards.
        """
        return self._cards


__all__ = ["Deck"]


"""Console script for bulkdata. (Currently does nothing)"""
import sys
import click


@click.command()
def main(args=None):
    """Console script for bulkdata."""
    click.echo("CLI not available yet...")
    return 0


if __name__ == "__main__":
    sys.exit(main())  # pragma: no cover



import re

try:
    from pyNastran.bdf.field_writer_8 import print_field_8
    from pyNastran.bdf.field_writer_16 import print_field_16

except ImportError:
    from .pyNastran.bdf.field_writer_8 import print_field_8
    from .pyNastran.bdf.field_writer_16 import print_field_16


rx_REAL_pat = r"""
    [-+]? # optional sign
    (?:
        (?: \d* \. \d+ ) # .1 .12 .123 etc 9.1 etc 98.1 etc
        |
        (?: \d+ \.? ) # 1. 12. 123. etc 1 12 123 etc
    )
    # followed by optional exponent part if desired
    (?: 
        (?: [Ee]? [+-]?) \d+ 
    ) 
    ?
"""
rx_REAL = re.compile(rx_REAL_pat, re.VERBOSE)

rx_INT = re.compile(r"^[-+]?([1-9]\d*|0)$")


def _force_E(real_field):
    # if no "E" or "e", insert it
    return re.sub("(?<!^)(?<![E|e])[+-]", r"E\g<0>", real_field)


def _is_match(rx, field):
    field = field.strip()
    try:
        return rx.match(field).group(0) == field
    except (TypeError, IndexError, AttributeError):
        return False


def is_integer_field(field):
    return _is_match(rx_INT, field)


def is_real_field(field):
    return _is_match(rx_REAL, field)


def read_integer_field(field):
    return int(field)


def read_real_field(field):
    return float(_force_E(field))


def read_field(field):
    """Convert `field` string to value
    """
    if is_integer_field(field):
        return read_integer_field(field)
    if is_real_field(field):
        return read_real_field(field)
    else:
        return field.strip()
    

def write_field(value, fieldspan=1):
    """Convert `value` to field string
    """
    width = fieldspan * 8
    if isinstance(value, str):
        return value[:width].strip()
    elif width == 8:
        return print_field_8(value).strip()
    elif width == 16:
        return print_field_16(value).strip()
    else:
        raise ValueError("non-character field entries "
                         "cannot span more than 2 fields, "
                         "but fieldspan is {}".format(fieldspan))


class Field:
    
    def __init__(self, value, fieldspan=1):
        self.span = fieldspan
        self.value = value

    @property
    def width(self):
        return self.span * 8
    
    def __eq__(self, other):
        if isinstance(other, Field):
            return self.raw == other.raw
        elif isinstance(other, str):
            return self.raw == other
        else:
            return self.value == other

    def __str__(self):
        return self.raw
    
    def __repr__(self):
        return "{}('{}')".format(self.__class__.__name__, self.raw)

    def __bool__(self):
        return bool(self.raw)
    
    @property
    def value(self):
        if self._value_change:
            self._value = read_field(self.raw)
        return self._value
    
    @value.setter
    def value(self, new_value):
        self.raw = write_field(new_value, fieldspan=self.span)
        self._value_change = True
        
    
class LargeField(Field):
    
    def __init__(self, value, fieldspan=2):
        super().__init__(value, fieldspan)
        
    def split(self, fieldspan=1):
        """Split into Fields with given `fieldspan`
        """
        fieldwidth = fieldspan * 8
        fields = []
        for start in range(0, len(self.raw), fieldwidth):
            stop = start + fieldwidth
            fields.append(Field(self.raw[start:stop], fieldspan))
        return fields

    @classmethod
    def join(cls, fields):
        large_str = "".join([field.raw for field in fields])
        return cls(large_str, fieldspan=len(fields))


"""Top-level package for Bulk Data Python Package."""

__author__ = """Marco T. de Lannoy Kobayashi"""
__email__ = 'mdlkdev@gmail.com'
__version__ = '0.5.2'


from .deck import Deck
from .card import Card


"""The :mod:`~bulkdata.card` module provides classes related
to the creation and manipulation of bulk data card objects.
It contains the :class:`~bulkdata.card.Card` class that 
facilitates the conversion of collections of field values to
and from bulk data card formatted strings.
"""

from collections import OrderedDict

from .field import Field, LargeField
from .format import format_card
from .parse import BDFParser
from .util import islist, split_fields, repr_list


class Card:
    """:class:`~bulkdata.card.Card` class allows the user 
    to create and modify bulk data cards.

    :param name: The name of the card
    :param size: The number of initial blank fields, defaults to 0
    """
    
    def __init__(self, name=None, size=0):
        self.name = name
        self._fields = [self._blank_field() for _ in range(size)]

    def _convert_to_fields(self, value, fieldspan=1):
        if fieldspan == 1:
            return [Field(value)]
        elif fieldspan > 1:
            return LargeField(value, fieldspan).split()
        else:
            raise ValueError("fieldspan < 1")

    def _blank_field(self):
        return Field(None)

    def set_raw_fields(self, fields):
        """Set the fields directly, without internal conversion of `fields`
        values to `Field` objects.
        
        .. note:: 
            
            The user should avoid using this function unless he/she
            knows what they are doing.
        """
        self._fields = fields

    def append(self, value, fieldspan=1):
        """Append a field value to card fields.

        :param value: The field value
        :param fieldspan: The number of field cells the value spans,
                          defaults to 1
        """
        self._fields.extend(self._convert_to_fields(value, fieldspan))

    def extend(self, values, fieldspan=1):
        """Extend card fields with sequence of field values.

        :param values: The sequence of field values
        :param fieldspan: The number of field cells that each value spans,
                          defaults to 1
        """
        if fieldspan == 1:
            fields = [Field(value) for value in values]
            self._fields.extend(fields)
        else:
            for value in values:
                self._fields.extend(self._convert_to_fields(value, fieldspan))

    def pop(self):
        """Remove the last field.
        """
        return self._fields.pop()
        
    def resize(self, size):
        """Resize the card fields to contain *size* fields. If current
        number of fields is greater than *size*, the extra fields will
        be removed from the end. If number of fields is less than *size*,
        extra blank fields are appended.

        :param size: The desired number of fields
        """
        numfields = len(self._fields)
        diff = size - numfields
        if diff > 0:
            for _ in range(diff):
                self._fields.append(Field(None))
        if diff < 0:
            for _ in range(abs(diff)):
                self._fields.pop()
    
    def strip(self): #TODO: rename to rstrip ?
        """Remove any trailing blank fields.
        """
        for i in reversed(range(len(self._fields))):
            if not self._fields[i]:
                del self._fields[i]
            else:
                break
        
    def _setsinglefield(self, index, value):
        """Set a single field value at the index.
        """
        self._fields[index] = Field(value)
        
    def _setmultifieldlist(self, indexs, values):
        """Set list of field values at the given indexes.
        """
        numindexs = len(indexs)
        if len(values) > numindexs:
            raise IndexError("Number of values, {}, greater than number "
                             "of indexes, {}.\n"
                             "Values: {}\n"
                             "Indexs: {}"
                             .format(len(values), numindexs, values, indexs))
        for i, index in enumerate(indexs):
            try:
                value = values[i]
            except IndexError:
                # no more values, set blank
                value = None
            self._setsinglefield(index, value)
            
    def _setmultifieldvalue(self, indexs, value):
        """Set a field value spanning multiple field cells given 
        by indexes.
        """
        fieldspan = len(indexs)
        if fieldspan < 1:
            raise ValueError("fieldspan < 1")
        fields = self._convert_to_fields(value, fieldspan=fieldspan)
        for i, index in enumerate(indexs):
            try:
                new_field = fields[i]
            except IndexError:
                new_field = self._blank_field()
            self._fields[index] = new_field
    
    def _setmultifield(self, indexs, value):
        """Set field value(s) spanning multiple field cells.
        """
        if islist(value):
            self._setmultifieldlist(indexs, value)
        else:
            self._setmultifieldvalue(indexs, value)

    def _convert_slice_to_steps(self, slice_):
        """Convert a slice to a list of the corresponding steps.
        """
        start = slice_.start or 0
        stop  = slice_.stop or len(self._fields)
        step  = slice_.step or 1
        return list(range(start, stop, step))
        
    def __setitem__(self, key, value):
        """Set field item into the card.

        :type key: int, slice, list
        :param key: The indexing key denoting where to set 
                    the field(s)
        :param value: The field value(s) to set
        """
        if isinstance(key, int):
            self._setsinglefield(key, value)
        elif isinstance(key, slice):
            steps = self._convert_slice_to_steps(key)
            self._setmultifield(steps, value)
        elif islist(key):
            self._setmultifield(key, value)
        else:
            raise TypeError(key, type(key))

    def _getsinglefield(self, index):
        """Get a single field value at the index.
        """
        return self._fields[index].value

    def _getmultifield(self, indexs):
        """Get list of field values at the given indexes.
        """
        return [self._fields[i].value for i in indexs]

    def __getitem__(self, key):
        """Get field item from the card.

        :type key: int, slice, list
        :param key: The indexing key denoting which field(s) to get
        :return: The field value(s)
        """
        if isinstance(key, int):
            return self._getsinglefield(key)
        elif isinstance(key, slice):
            key = self._convert_slice_to_steps(key)
            return self._getmultifield(key)
        elif islist(key):
            return self._getmultifield(key)
        else:
            raise TypeError(key, type(key))

    def get_large(self, key):
        """Get field value spanning multiple fields. This gets the
        same fields as :meth:`~bulkdata.card.Card.__getitem__`
        but it joins them before converting to a single large value.
        
        :param key: The indexing object denoting which field(s) to get
        :return: The large field value
        """
        if isinstance(key, int):
            return self._getsinglefield(key)
        elif isinstance(key, slice):
            fields = self._fields[key]
        elif islist(key): 
            fields = [self._fields[i] for i in key]
        else:
            raise TypeError(key, type(key))
        return LargeField.join(fields).value

    def __delitem__(self, key):
        """Delete the field item from card.
        
        :type key: int, slice
        :param key: The indexing key denoting which field(s)
                    to delete
        """
        self._fields.__delitem__(key)
            
    def dumps(self, format="fixed"):
        """Dump the card to bulk data formatted string.

        :param format: the desired format, can be one of: 
                       ["free", "fixed"], defaults to "fixed"
        :return: The bulk data card string representation
        """
        return format_card(self, format)
    
    @classmethod
    def loads(cls, card_str):
        """Load :class:`~bulkdata.card.Card` object from a
        bulk data card string.

        :param card_str: the bulk data card string
        :return: The loaded :class:`~bulkdata.card.Card` object
        """
        card_name, card_fields = BDFParser(card_str).parse_card()
        obj = cls(card_name)
        obj.set_raw_fields([Field(value) for value in card_fields])
        
        return obj

    def values(self):
        """Get a list of the values of the card fields.
        """
        return [field.value for field in self._fields]
    
    def __contains__(self, value):
        """Return ``True`` if the card contains a field
        with the specified value, ``False`` otherwise.

        :param value: The specified value
        """
        return self._fields.__contains__(value)
    
    def __str__(self):
        """Dump the card to as a bulk data card string
        with default format.
        
        :return: The bulk data card string
        """
        return self.dumps()

    def __len__(self):
        """Return number of fields in the card.
        """
        return self.fields.__len__()

    def __iter__(self):
        """Iterate through the fields in the card.
        """
        return self.fields.__iter__()

    def __bool__(self):
        """Return ``True`` if the card contains any fields,
        ``False`` otherwise.
        """ 
        return bool(self.fields)

    def __repr__(self):
        return "{}(\"{}\", {})".format(self.__class__.__name__, 
                                   self.name, 
                                   repr_list(self.values()))

    @property
    def name(self):
        """The card name.
        """
        return self._name
    
    @name.setter
    def name(self, new_name):
        if new_name:
            new_name = new_name.strip()
        self._name = new_name
    
    @property
    def fields(self):
        """The card fields.
        """
        return self._fields


# class CardType:
    
#     def __init__(self, name):
#         self._name = name
#         self._entries = OrderedDict()

#     def register(self, name, index, default=None, valid=None, fieldspan=1):
#         entry_meta = {
#             "index": index,
#             "default": default,
#             "valid": valid,
#             "fieldspan": fieldspan
#         }
#         self._entries[name] = entry_meta

#     def _validate(self, valid, value):
#         if valid and not valid(value):
#             raise Exception("failed validation") #TODO

#     def __call__(self, *args, **kwargs):
        
#         entry_names = list(self._entries.keys())
#         entry_values = {}
        
#         for arg in args:
#             name = entry_names.pop(0)
#             entry_values[name] = arg
            
#         for name, arg in kwargs.items():
#             entry_names.pop(entry_names.index(name))
#             entry_values[name] = arg
            
#         # entries not handled by args or kwargs, try to use defaults
#         for name in entry_names:
#             default = self._entries[name].get("default")
#             if default is not None:
#                 entry_values[name] = default
#             else:
#                 raise ValueError("No value specified for required field: {}".format(name))

#         # validate and count total number of fields required
#         numfields = 0
#         for name, entry_meta in self._entries.items():
            
#             value = entry_values[name]
#             fieldspan = entry_meta["fieldspan"]
#             valid = entry_meta["valid"]
            
#             if islist(value):
                
#                 for each in value:
#                     self._validate(valid, each)
#                 numfields += len(value) * fieldspan
                
#             else:
                
#                 self._validate(valid, value)
#                 numfields += 1 * fieldspan
        
#         card = Card(self._name, size=numfields)
        
#         for name, entry_meta in self._entries.items():
            
#             value = entry_values[name]

#             card[entry_meta["index"]] = value
            
#         return card


__all__ = ["Card"]


from collections.abc import Sequence


def islist(value):
    return isinstance(value, Sequence) and not isinstance(value, str)


def repr_list(list_):
    if len(list_) > 10:
        return "[{} ... {}]".format(repr(list_[0]), repr(list_[-1]))
    else:
        return repr(list_)


def split_fields(fields_str, fieldwidth=8):
    """Split `fields_str` into fields of length `fieldwidth`
    """
    fields = []
    for start in range(0, len(fields_str), fieldwidth):
        stop = start + fieldwidth
        fields.append(fields_str[start:stop])
    return fields


#!/usr/bin/env python

"""Tests for `bulkdata.field` module."""

import pytest

from bulkdata.field import Field, LargeField
from bulkdata.field import _force_E
from bulkdata.field import is_integer_field, is_real_field
from bulkdata.field import read_integer_field, read_real_field, read_field
from bulkdata.field import write_field


@pytest.fixture
def integer_fields():
    return ("0", "1", "10000000", "-9999999")


@pytest.fixture
def integer_values():
    return (0, 1, 10000000, -9999999)


@pytest.fixture
def real_fields():
    return ("3.14+10", "3.14-10", "0.e+19", "0.e-19", "3.14", "-3.14")


@pytest.fixture
def real_values():
    return (3.14e10, 3.14e-10, 0.e19, 0.e-19, 3.14, -3.14)


def test_force_E():

    assert _force_E("3.14") == "3.14"
    assert _force_E("3.14+10") == "3.14E+10"
    assert _force_E("3.14-10") == "3.14E-10"


def test_is_integer_field(integer_fields, real_fields):

    for field in integer_fields:
        assert is_integer_field(field)

    for field in real_fields:
        assert not is_integer_field(field)


def test_is_real_field(real_fields):

    for field in real_fields:
        assert is_real_field(field)
    

def test_read_integer_field(integer_fields, integer_values):

    for field, value in zip(integer_fields, integer_values):
        assert read_integer_field(field) == value
        assert read_field(field) == value
    

def test_read_real_field(real_fields, real_values):

    for field, value in zip(real_fields, real_values):
        assert read_real_field(field) == value
        assert read_field(field) == value

def test_read_field_integer():

    test_tuples = [
        ("       1", 1),
        ("      12", 12),
        ("     123", 123),
        ("    1234", 1234),
        ("   12345", 12345),
        ("  123456", 123456),
        (" 1234567", 1234567),
        ("12345678", 12345678),
    ]
    
    for integer_str, expect in test_tuples:
        field = read_field(integer_str)
        assert field == expect

    test_long = [
        ("       -12345678", -12345678),
        ("       123456789", 123456789)
    ]
    
    for integer_str, expect in test_long:
        field = read_field(integer_str)
        assert field == expect


def test_read_field_real():

    test_tuples = [
        ("1234567.", 1234567.), 
        ("-123456.", -123456.), 
        ("1.235+13", 1.235e13), 
        ("123.4567", 123.4567), 
        ("123.4568", 123.4568), 
        ("1.234568", 1.234568), 
        ("0.123457", 0.123457), 
        ("      1.", 1.), 
        ("     12.", 12.), 
        ("    123.", 123.), 
        ("   1234.", 1234.), 
        ("  12345.", 12345.), 
        (" 123456.", 123456.), 
        ("1.2346+7", 1.2346e7), 
        ("-1.235+6", -1.235e6),
        ("9233443.", 9233443.)
    ]
    
    for real_str, expect in test_tuples:
        field = read_field(real_str)
        assert field == expect


def test_write_integer_field():

    test_tuples = [
        ("1", 1),
        ("12", 12),
        ("123", 123),
        ("1234", 1234),
        ("12345", 12345),
        ("123456", 123456),
        ("1234567", 1234567),
        ("12345678", 12345678),
    ]
    
    for expect, integer in test_tuples:
        field = write_field(integer)
        assert field == expect

    test_long = [
        ("-12345678", -12345678),
        ("123456789", 123456789)
    ]
    
    for expect, integer in test_long:
        field = write_field(integer, fieldspan=2)
        assert field == expect


def test_write_integer_field_fail():

    test_fails = [
        -12345678,
        123456789
    ]
    
    for integer in test_fails:
        with pytest.raises(RuntimeError):
            write_field(integer)

    with pytest.raises(ValueError):
        write_field(500, fieldspan=3)


def test_write_real_field():

    test_tuples = [
        ("1234567.", 1234567.), 
        ("-123456.", -123456.), 
        ("1.235+13", 12345678901234.), 
        ("123.4567", 123.4567), 
        ("123.4568", 123.45678), 
        ("1.234568", 1.2345678), 
        (".1234568", 0.12345678), 
        ("1.", 1.), 
        ("12.", 12.), 
        ("123.", 123.), 
        ("1234.", 1234.), 
        ("12345.", 12345.), 
        ("123456.", 123456.), 
        ("1.2346+7", 12345678.), 
        ("-1.235+6", -1234567.),
        ("9233443.", 9233443.323)
    ]
    
    for expect, real in test_tuples:
        field = write_field(real)
        assert field == expect


def test_write_real_field_fail():

    with pytest.raises(ValueError):
        write_field(500., fieldspan=3)


def test_write_string_field():

    test_tuples = [
        ("this", "this    "),
        ("is", "      is"),
        ("a-test", "a-test"),
        ("hellowor", "helloworld")
    ]

    for expect, value in test_tuples:
        field = write_field(value)
        assert field == expect


def test_write_string_largefield():

    test_tuples = [
        ("helloworld", 2),
        ("the-answer-to-life-the-universe-and-everything", 6)
    ]

    for value, fieldspan in test_tuples:
        assert write_field(value, fieldspan) == value


#!/usr/bin/env python

"""Tests for `bulkdata.parse` module."""

import pytest

from bulkdata.parse import BDFParser

from . import BDF_DIR, EXPECT_DIR


def test_parse_card_simple():

    card_str = """\
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2       +0      
+0      9.9
$
$
"""

    name, fields = BDFParser(card_str).parse_card()
    assert name.strip() == "HELLO"
    assert len(fields) == 9


def test_parse_card_mixedcont():
    
    card_str = """\
BARL    6666    10              BAR
+       5.5     3.0
"""

    name, fields = BDFParser(card_str).parse_card()
    assert name.strip() == "BARL"
    assert len(fields) == 10


def test_parse_card_sparse():

    card_str = """\
SPARSE  test1
        test2
        test3
"""
    name, fields = BDFParser(card_str).parse_card()
    assert name.strip() == "SPARSE"
    assert len(fields) == 17


def test_parse_deck_cards_only():

    deck_str = """\
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2       +0      
+0      9.9
$
$
HELLO,99,hellowor,ld,0,3.3,1,6.6,2,+0      
+0,9.9
$
$
HELLO,99,hellowor,ld,0,3.3,1,6.6,2,
,9.9
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2       +0      
+0      9.9
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2       +0      
+0      9.9
$
$
"""

    header, card_tuples = BDFParser(deck_str).parse()

    assert not header
    assert len(card_tuples) == 5
    for name, fields in card_tuples:
        assert name.strip() == "HELLO"
        assert len(fields) == 9


def test_parse_deck():

    header_str = """\
THIS
IS
A
TEST
ANSWER IS 42\
"""
    deck_str = """{}
BEGIN BULK
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2       +0      
+0      9.9
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2       +0      
+0      9.9
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2       +0      
+0      9.9
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2       +
+       9.9
$
$
HELLO   99      helloworld      0       3.3     1       6.6     2
        9.9
$
$
ENDDATA
IGNORE THIS
""".format(header_str)

    header, card_tuples = BDFParser(deck_str).parse()

    assert header == header_str
    assert len(card_tuples) == 5
    for name, fields in card_tuples:
        assert name.strip() == "HELLO"
        assert len(fields) == 9


def test_parse_bdf_pyNastran():

    bdf_filename = BDF_DIR + "/testA.bdf"
    expect_header_filename = EXPECT_DIR + "/testA-header.bdf"

    with open(bdf_filename) as bdf_file:
        bdf_str = bdf_file.read()

    with open(expect_header_filename) as header_file:
        expect_header = header_file.read()

    header, card_tuples = BDFParser(bdf_str).parse()

    assert header == expect_header
    assert len(card_tuples) == 143



"""
defines:
 - fields_out = build_table_lines(fields, nstart=1, nend=0)
 - fields_out = wipe_empty_fields(card)

"""
from typing import List, Union


def build_table_lines(fields, nstart=1, nend=0):
    """
    Builds a table of the form:

    +--------+-------+--------+-------+--------+-------+-------+-------+
    | DESVAR | DVID1 | DVID2  | DVID3 |DVID4   | DVID5 | DVID6 | DVID7 |
    +--------+-------+--------+-------+--------+-------+-------+-------+
    |        | DVID8 |  etc.  |       |        |       |       |       |
    +--------+-------+--------+-------+--------+-------+-------+-------+
    |        |  UM   | VAL1   | VAL2  |  etc.  |       |       |       |
    +--------+-------+--------+-------+--------+-------+-------+-------+

    and pads the rest of the fields with None's (e.g. at the end of the
    DVID8 line).

    Parameters
    ----------
    fields: List[int/float/str]
        the fields to enter, including DESVAR
    nstart: int; default=1
        the number of blank fields at the start of the line
    nend : int; default=0)
        the number of blank fields at the end of the line

    Returns
    -------
    fields_out : List[str/None]
        the values ready for card printing

    .. note:: will be used for DVCREL2, DVMREL2, DVPREL2, RBE1, RBE3, DRESP2, DRESP3
    .. warning:: only works for small field format???
    """
    fields_out = []
    n = 8 - nstart - nend

    # pack all the fields into a list.  Only consider an entry as isolated
    for (i, field) in enumerate(fields):
        fields_out.append(field)
        if i > 0 and i % n == 0:  # beginning of line
            #pad = [None] * (i + j)
            #fields_out += pad
            fields_out += [None] * (nstart + nend)

    # make sure they're aren't any trailing None's (from a new line)
    fields_out = wipe_empty_fields(fields_out)

    # push the next key (aka next fields[0]) onto the next line
    nspaces = 8 - (len(fields_out)) % 8  # puts UM onto next line
    if nspaces < 8:
        fields_out += [None] * nspaces
    return fields_out


def wipe_empty_fields(card: List[Union[str, int, float, None]]) -> List[Union[str, int, float, None]]:
    """
    Removes any trailing Nones from the card.
    Also converts empty strings to None.
    Allows floats & ints, but that's not the intended value,
    though it is ok (it's just extra work).

    Parameters
    ----------
    card : List[str]
        the fields on the card as a list

    Returns
    -------
    short_card : List[str]
        the card with no trailing blank fields

    """
    short_card = []  # type: List[Union[str, int, float, None]]
    for field in card:
        if isinstance(field, str):
            field = field.strip()
            if field == '':
                short_card.append(None)
            else:
                short_card.append(field)
        else:
            short_card.append(field)

    imax = 0
    for i, field in enumerate(card):
        if short_card[i] is not None:
            imax = i
    out = short_card[:imax + 1]
    return out



"""Interface to various numpy utilities"""
import numpy as np

integer_types = (int, np.int32, np.int64)
integer_string_types = (int, np.int32, np.int64, bytes, str)
integer_float_types = (int, np.int32, np.int64, float, np.float32)
float_types = (float, np.float32, np.float64)
bytes_type = (bytes, np.bytes_)



from .field import Field


class BaseFormatter:

    newline = "\n"
    delimiter = ""
    valuesperline = 8 # fields with value, per line
    fieldwidth = 8

    def format_field(self, field):
        if isinstance(field, Field):
            field = field.raw
        else:
            field = field
        return field[:self.fieldwidth]
        
    def continuation(self, index=""):
        label = "+" + str(index)
        beg = self.format_field(label)
        end = self.format_field(label)
        return beg + self.newline + end

    def endofline(self, index):
        valuesperline = self.valuesperline
        if index and valuesperline:
            if index < valuesperline:
                return False
            else:
                return True
        else:
            return False

    def format_card(self, card):
        
        card_str = ""
        line_pos = 0   # position in the line
        line_count = 0 # line counter
        
        def concat_field(field, last_field=False):
            nonlocal card_str, line_pos, line_count

            card_str += self.format_field(field or " ")
            card_str += self.delimiter
            line_pos += 1
            
            # if end-of-line is reached, concat continuation
            if self.endofline(line_pos) and not last_field:
                card_str += self.continuation(line_count)
                card_str += self.delimiter
                line_pos = 0
                line_count += 1

        fields = card.fields
        card_str += self.format_field(card.name or " ")

        if len(fields) > 0:
            card_str += self.delimiter

            for field in fields[:-1]:
                concat_field(field)
            
            concat_field(fields[-1], last_field=True)
        
        return card_str.rstrip(" " + self.delimiter) + self.newline


class FixedFormatter(BaseFormatter):

    def format_field(self, field):
        field = super().format_field(field)
        return "{:<{}}".format(field, self.fieldwidth)


class FreeFormatter(BaseFormatter):

    delimiter = ","


_formatters = {
    "fixed": FixedFormatter(),
    "free": FreeFormatter()
}

_defaultformat = "fixed"


def format_card(card, format=None):

    format = format or _defaultformat

    return _formatters[format].format_card(card)



#!/usr/bin/env python

"""Tests for `bulkdata.deck` module."""

import pytest

# from bulkdata.format import FixedFormat, FreeFormat
from bulkdata.card import Card
from bulkdata.deck import Deck

from . import BDF_DIR, EXPECT_DIR

@pytest.fixture 
def cards():
    
    def make_card(name, id, keys, integers, reals):
        card = Card(name)
        card.append(id)
        for key, integer, real in zip(keys, integers, reals):
            card.extend([key, integer, real])
        return card

    cards_name = ("ONE", "TWO", "THREE", "FOUR")
    cards_id = (
        100 + i
        for i in range(len(cards_name))
    )
    cards_keys = (
        ("this-%d" % i, "that-%d" % i, "other-%d" % i) 
        for i in range(len(cards_name))
    )
    cards_integers = (
        (1 * i, 2 * i, 3 * i) 
        for i in range(len(cards_name))
    )
    cards_reals = (
        (2.1 * i, 4.2 * i, 8.4 * i) 
        for i in range(len(cards_name))
    )
    cards = [
        make_card(name, id, keys, integers, reals)
        for name, id, keys, integers, reals in zip(cards_name,
                                                   cards_id,
                                                   cards_keys,
                                                   cards_integers,
                                                   cards_reals)
    ]
    return cards


def test_deck_init_iter(cards):

    deck = Deck(cards)
    assert len(deck) == len(cards)

    for i, card in enumerate(deck):
        assert card == cards[i]


def test_deck_append(cards):

    deck = Deck()
    assert len(deck) == 0

    for card in cards:
        deck.append(card)
    assert len(deck) == len(cards)

    for i, card in enumerate(deck):
        assert card == cards[i]


def test_deck_extend(cards):

    deck = Deck()
    assert len(deck) == 0

    deck.extend(cards)
    assert len(deck) == len(cards)

    for i, card in enumerate(deck):
        assert card == cards[i]


def test_deck_find(cards):

    deck = Deck(cards)

    found = list(deck.find())
    assert len(found) == len(cards)

    for name in ("ONE", "TWO", "THREE", "FOUR"):
        found = list(deck.find({"name": name}))
        found2 = list(deck.find(name))
        assert len(found) == 1
        assert len(found2) == 1


def test_deck_find_fields(cards):
    pass


def test_deck_find_contains(cards):
    pass


def test_deck_find_one():

    deck = Deck()

    for _ in range(10):
        deck.append(Card("CLONE"))

    card = deck.find_one("CLONE")
    assert card.name == "CLONE"
    assert card == list(deck.find("CLONE"))[0]


def test_deck_find_notexist():
   
    deck = Deck()

    for _ in range(10):
        deck.append(Card("CLONE"))

    card = deck.find_one("UNIQUE")
    assert card is None

    cards = list(deck.find("UNIQUE"))
    assert not cards


def test_deck_replace(cards):

    deck = Deck(cards)

    replacement = Card("UNO")
    deck.replace({"name": "ONE"}, replacement)

    assert not list(deck.find({"name": "ONE"}))
    assert list(deck.find({"name": "UNO"}))
    assert len(deck) == len(cards)


def test_deck_update(cards):

    deck = Deck(cards)

    index = 0
    index_value = "int"
    update = {"index": index, "value": index_value}
    deck.update({"name": "ONE"}, update)

    card = next(deck.find({"name": "ONE"}))
    assert card[index] == index_value

    slice_ = slice(0, 2)
    slice_value = "sliceslice"
    update = {"index": slice_, "value": slice_value}
    deck.update({"name": "TWO"}, update)

    card = next(deck.find({"name": "TWO"}))
    assert card[:2] == ["slicesli", "ce"]
    assert card.get_large(slice_) == slice_value

    indexs =[-4, -3, -2, -1]
    indexs_values = [1001, 1002, 1003, 1004]
    update = {"index": indexs, "value": indexs_values}
    deck.update({"name": "THREE"}, update)

    card = next(deck.find({"name": "THREE"}))
    assert card[indexs] == indexs_values

    for notlisttype in (slice, int, str):
        with pytest.raises(TypeError):
            deck.update({}, {"index": notlisttype(), "value": list()})

    deck_str = deck.dumps()

    # redo the update, but without using update func
    deck2 = Deck(cards)

    card = deck.find_one({"name": "ONE"})
    card[index] = index_value 

    card = deck.find_one({"name": "TWO"})
    card[slice_] = slice_value

    card = deck.find_one({"name": "THREE"})
    card[indexs] = indexs_values

    deck2_str = deck2.dumps()

    assert deck_str == deck2_str


def test_deck_load_bdf_pyNastran():

    bdf_filename = BDF_DIR + "/testA.bdf"

    with open(bdf_filename) as bdf_file:
        deck = Deck.load(bdf_file)

    aero = deck.find_one("AERO")
    assert not (aero[0] or aero[1])
    assert aero[2] == aero[3] == 1.0
    
    assert not deck.find_one({"name": None})
    assert not deck.find_one({"name": ""})
    assert not deck.find_one({"name": "+"})

    with open(EXPECT_DIR + "/testA-fixed.bdf") as f:
        assert deck.dumps("fixed") == f.read()

    with open(EXPECT_DIR + "/testA-free.bdf") as f:
        assert deck.dumps("free") == f.read()


def test_zaero_example():

    bdf_filename = BDF_DIR + "/zaero-example.bdf"

    # load Deck from BDF file
    with open(bdf_filename) as bdf_file:
        deck = Deck.load(bdf_file)

    # CORD2R variables
    cid = 1
    rid = None
    a = [-2.9, 1.0, 0.0]
    b = [3.6, 0.0, 1.0]
    c = [5.2, 1.0, -2.9]

    # create CORD2R card
    cord2r = Card("CORD2R")
    cord2r.append(cid)
    cord2r.append(rid)
    cord2r.extend(a)
    cord2r.extend(b)
    cord2r.extend(c)

    # # print the CORD2R card in fixed format (the default)
    # print("-- CORD2R fixed formatting --")
    # print(cord2r.dumps("fixed"))

    # # print the CORD2R card in free format
    # print("-- CORD2R free formatting --")
    # print(cord2r.dumps("free"))

    # add card to the deck
    deck.append(cord2r)

    # get AEROZ card
    aeroz = deck.find_one({"name": "AEROZ"})

    # print("-- AEROZ before update --")
    # print(aeroz.dumps())

    # update the ACSID field (first one)
    aeroz[0] = cid

    # update mass and length units fields while we're at it
    aeroz[[3, 4]] = ["N", "M"] 

    # print("-- AEROZ after update --")
    # print(aeroz.dumps())

    # dump Deck to update BDF file
    with open(EXPECT_DIR + "/zaero-example-update.bdf") as f:
        assert deck.dumps() == f.read()

    # with open(EXPECT_DIR + "/zaero-example-update.bdf", "w") as f:
    #     deck.dump(f)
    # assert False


    




    


#!/usr/bin/env python

"""Tests for `bulkdata.format` module."""

import pytest

from bulkdata.format import FixedFormatter, FreeFormatter

from .util import MockCard


@pytest.fixture
def fixedform():
    return FixedFormatter()


@pytest.fixture
def freeform():
    return FreeFormatter()


def test_write_string_field_fixed(fixedform):

    field = fixedform.format_field("hello")
    assert field == "hello   "

    field = fixedform.format_field("helloworld")
    assert field == "hellowor"


def test_write_string_field_free(freeform):

    field = freeform.format_field("hello")
    assert field == "hello"

    field = freeform.format_field("helloworld")
    assert field == "hellowor"


def test_format_card_fixed(fixedform):

    name = "TEST"
    fields_str = "yes these fields are not properly formatted this is a test"
    fields = fields_str.split()
    card = MockCard(name, fields)
    actual = fixedform.format_card(card)
    expect = """\
TEST    yes     these   fields  are     not     properlyformattethis    +0      
+0      is      a       test
"""
    assert actual == expect


def test_format_card_free(freeform):

    name = "TEST"
    fields_str = "yes these fields are not properly formatted this is a test"
    fields = fields_str.split()
    card = MockCard(name, fields)
    actual = freeform.format_card(card)
    expect = """\
TEST,yes,these,fields,are,not,properly,formatte,this,+0
+0,is,a,test
"""
    assert actual == expect


def test_write_empty_card(fixedform, freeform):

    card = MockCard(None, [])

    for formatter in (fixedform, freeform):
        actual = formatter.format_card(card)
        expect = "\n"
        assert actual == expect




class MockCard():

    def __init__(self, name, fields):
        self._name = name
        self._fields = fields

    @property
    def name(self):
        return self._name

    @property
    def fields(self):
        return self._fields